// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const EncapTable = "Encap"

type (
	EncapType = string
)

var (
	EncapTypeGeneve EncapType = "geneve"
	EncapTypeSTT    EncapType = "stt"
	EncapTypeVxlan  EncapType = "vxlan"
)

// Encap defines an object in Encap table
type Encap struct {
	UUID        string            `ovsdb:"_uuid"`
	ChassisName string            `ovsdb:"chassis_name"`
	IP          string            `ovsdb:"ip"`
	Options     map[string]string `ovsdb:"options"`
	Type        EncapType         `ovsdb:"type" validate:"oneof='geneve' 'stt' 'vxlan'"`
}
