// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const ControllerEventTable = "Controller_Event"

type (
	ControllerEventEventType = string
)

var ControllerEventEventTypeEmptyLbBackends ControllerEventEventType = "empty_lb_backends"

// ControllerEvent defines an object in Controller_Event table
type ControllerEvent struct {
	UUID      string                   `ovsdb:"_uuid"`
	Chassis   *string                  `ovsdb:"chassis"`
	EventInfo map[string]string        `ovsdb:"event_info"`
	EventType ControllerEventEventType `ovsdb:"event_type" validate:"oneof='empty_lb_backends'"`
	SeqNum    int                      `ovsdb:"seq_num"`
}
