package qemudefault

// CPUCores defines the default number of cores a VM will get if no limit specified.
const CPUCores = 1

// MemSize is the default memory size for VMs if no limit specified.
const MemSize = "1GiB"
