// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const RouteTable = "Route"

type (
	RouteOrigin = string
)

var (
	RouteOriginConnected RouteOrigin = "connected"
	RouteOriginStatic    RouteOrigin = "static"
)

// Route defines an object in Route table
type Route struct {
	UUID             string            `ovsdb:"_uuid"`
	AvailabilityZone string            `ovsdb:"availability_zone"`
	ExternalIDs      map[string]string `ovsdb:"external_ids"`
	IPPrefix         string            `ovsdb:"ip_prefix"`
	Nexthop          string            `ovsdb:"nexthop"`
	Origin           RouteOrigin       `ovsdb:"origin" validate:"oneof='connected' 'static'"`
	RouteTable       string            `ovsdb:"route_table"`
	TransitSwitch    string            `ovsdb:"transit_switch"`
}
