// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const RBACPermissionTable = "RBAC_Permission"

// RBACPermission defines an object in RBAC_Permission table
type RBACPermission struct {
	UUID          string   `ovsdb:"_uuid"`
	Authorization []string `ovsdb:"authorization"`
	InsertDelete  bool     `ovsdb:"insert_delete"`
	Table         string   `ovsdb:"table"`
	Update        []string `ovsdb:"update"`
}
