// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const DHCPOptionsTable = "DHCP_Options"

type (
	DHCPOptionsType = string
)

var (
	DHCPOptionsTypeBool         DHCPOptionsType = "bool"
	DHCPOptionsTypeUint8        DHCPOptionsType = "uint8"
	DHCPOptionsTypeUint16       DHCPOptionsType = "uint16"
	DHCPOptionsTypeUint32       DHCPOptionsType = "uint32"
	DHCPOptionsTypeIpv4         DHCPOptionsType = "ipv4"
	DHCPOptionsTypeStaticRoutes DHCPOptionsType = "static_routes"
	DHCPOptionsTypeStr          DHCPOptionsType = "str"
	DHCPOptionsTypeHostID       DHCPOptionsType = "host_id"
	DHCPOptionsTypeDomains      DHCPOptionsType = "domains"
)

// DHCPOptions defines an object in DHCP_Options table
type DHCPOptions struct {
	UUID string          `ovsdb:"_uuid"`
	Code int             `ovsdb:"code" validate:"min=0,max=254"`
	Name string          `ovsdb:"name"`
	Type DHCPOptionsType `ovsdb:"type" validate:"oneof='bool' 'uint8' 'uint16' 'uint32' 'ipv4' 'static_routes' 'str' 'host_id' 'domains'"`
}
