// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const DNSTable = "DNS"

// DNS defines an object in DNS table
type DNS struct {
	UUID        string            `ovsdb:"_uuid"`
	ExternalIDs map[string]string `ovsdb:"external_ids"`
	Records     map[string]string `ovsdb:"records"`
}
