package daemon

// Debug indicates if daemon is started with debug mode.
var Debug bool

// Verbose indicates if daemon is started with verbose mode.
var Verbose bool

// SharedMountsSetup indicates if daemon has setup shared mounts.
var SharedMountsSetup bool
